@@ -8,13 +8,13 @@ from polars import selectors as cs
 from polars_ds import weighted_mean
 
 from value_dashboard.metrics.clv import rfm_summary
-from value_dashboard.metrics.constants import MODELCONTROLGROUP
-from value_dashboard.metrics.ml import binary_metrics_tdigest
+from value_dashboard.metrics.constants import MODELCONTROLGROUP, PROPENSITY
+from value_dashboard.metrics.ml import binary_metrics_tdigest, calibration_tdigest
 from value_dashboard.utils.config import get_config
 from value_dashboard.utils.logger import get_logger
 from value_dashboard.utils.polars_utils import merge_digests, estimate_quantile
+from value_dashboard.utils.py_utils import strtobool
 from value_dashboard.utils.stats import chi2_test, g_test, z_test, proportions_ztest
-from value_dashboard.utils.string_utils import strtobool
 from value_dashboard.utils.timer import timed
 
 logger = get_logger(__name__)
@@ -46,6 +46,8 @@ def group_model_ml_scores_data(
 ) -> pl.DataFrame:
     if isinstance(model_roc_auc_data, pd.DataFrame):
         model_roc_auc_data = pl.from_pandas(model_roc_auc_data)
+    if isinstance(model_roc_auc_data, pl.DataFrame):
+        model_roc_auc_data = model_roc_auc_data.clone()
 
     auc_data = model_roc_auc_data.filter(pl.col("Count") > 0)
 
@@ -53,7 +55,7 @@ def group_model_ml_scores_data(
     use_t_digest = (
         strtobool(m_config["use_t_digest"])
         if "use_t_digest" in m_config.keys()
-        else False
+        else True
     )
     logger.debug("Use t-digest for scores: " + str(use_t_digest))
 
@@ -114,6 +116,8 @@ def group_experiment_data(
 ) -> pl.DataFrame:
     if isinstance(exp_data, pd.DataFrame):
         exp_data = pl.from_pandas(exp_data)
+    if isinstance(exp_data, pl.DataFrame):
+        exp_data = exp_data.clone()
 
     m_config = get_config()["metrics"][config["metric"]]
     grp_by = (
@@ -144,6 +148,8 @@ def calculate_experiment_scores(
 ) -> pl.DataFrame:
     if isinstance(exp_data, pd.DataFrame):
         exp_data = pl.from_pandas(exp_data)
+    if isinstance(exp_data, pl.DataFrame):
+        exp_data = exp_data.clone()
     sort_list = []
     if "facet_row" in config.keys():
         sort_list.append(config["facet_row"])
@@ -175,19 +181,27 @@ def calculate_experiment_scores(
                     pl.map_groups(
                         exprs=[m_config["experiment_group"], "Positives", "Negatives"],
                         function=chi2_test,
-                        return_dtype=pl.Struct,
+                        return_dtype=pl.Struct([pl.Field("chi2_stat", pl.Float64), pl.Field("chi2_dof", pl.Int64),
+                                                pl.Field("chi2_p_val", pl.Float64),
+                                                pl.Field("chi2_odds_ratio_stat", pl.Float64),
+                                                pl.Field("chi2_odds_ratio_ci_low", pl.Float64),
+                                                pl.Field("chi2_odds_ratio_ci_high", pl.Float64)]),
                         returns_scalar=True,
                     ).alias("chi2_stat"),
                     pl.map_groups(
                         exprs=[m_config["experiment_group"], "Positives", "Negatives"],
                         function=g_test,
-                        return_dtype=pl.Struct,
+                        return_dtype=pl.Struct([pl.Field("g_stat", pl.Float64), pl.Field("g_dof", pl.Int64),
+                                                pl.Field("g_p_val", pl.Float64),
+                                                pl.Field("g_odds_ratio_stat", pl.Float64),
+                                                pl.Field("g_odds_ratio_ci_low", pl.Float64),
+                                                pl.Field("g_odds_ratio_ci_high", pl.Float64)]),
                         returns_scalar=True,
                     ).alias("g_stat"),
                     pl.map_groups(
                         exprs=[m_config["experiment_group"], "Positives", "Count"],
                         function=z_test,
-                        return_dtype=pl.Struct,
+                        return_dtype=pl.Struct([pl.Field("z_score", pl.Float64), pl.Field("z_p_val", pl.Float64)]),
                         returns_scalar=True,
                     ).alias("z_stat"),
                     pl.col("Count").sum(),
@@ -211,17 +225,26 @@ def calculate_model_ml_scores(
 ) -> pl.DataFrame:
     if isinstance(model_roc_auc_data, pd.DataFrame):
         model_roc_auc_data = pl.from_pandas(model_roc_auc_data)
+    if isinstance(model_roc_auc_data, pl.DataFrame):
+        model_roc_auc_data = model_roc_auc_data.clone()
 
-    grp_by = config["group_by"]
+    grp_by = config.get("group_by", None)
+    if not grp_by:
+        grp_by = ['ALL']
+        model_roc_auc_data = model_roc_auc_data.with_columns(
+            pl.lit('ALL').alias('ALL')
+        )
     m_config = get_config()["metrics"][config["metric"]]
     scores = m_config["scores"]
     use_t_digest = (
         strtobool(m_config["use_t_digest"])
         if "use_t_digest" in m_config.keys()
-        else False
+        else True
     )
     logger.debug("Use t-digest for scores: " + str(use_t_digest))
 
+    property = 'tdigest' if config.get('property', PROPENSITY) == PROPENSITY else 'tdigest_finalprop'
+
     auc_data = (
         model_roc_auc_data
         .group_by(grp_by)
@@ -242,27 +265,48 @@ def calculate_model_ml_scores(
             + (
                 [
                     pl.map_groups(
-                        exprs=["tdigest_positives",
-                               "tdigest_negatives"],
+                        exprs=[f"{property}_positives",
+                               f"{property}_negatives"],
                         function=binary_metrics_tdigest,
-                        return_dtype=pl.Struct,
+                        return_dtype=pl.Struct(
+                            [pl.Field("roc_auc", pl.Float64), pl.Field("average_precision", pl.Float64),
+                             pl.Field("tpr", pl.List(pl.Float64)), pl.Field("fpr", pl.List(pl.Float64)),
+                             pl.Field("precision", pl.List(pl.Float64)), pl.Field("recall", pl.List(pl.Float64)),
+                             pl.Field("pos_fraction", pl.Float64)]),
                         returns_scalar=True,
                     ).alias("roc_auc_tdigest_a")
                 ]
                 if use_t_digest
                 else []
             )
+            + (
+                [
+                    pl.map_groups(
+                        exprs=[f"{property}_positives",
+                               f"{property}_negatives"],
+                        function=calibration_tdigest,
+                        return_dtype=pl.Struct([pl.Field("calibration_bin", pl.List(pl.Float64)),
+                                                pl.Field("calibration_proba", pl.List(pl.Float64)),
+                                                pl.Field("calibration_rate", pl.List(pl.Float64))]),
+                        returns_scalar=True,
+                    ).alias("calibration_tdigest_a")
+                ]
+                if use_t_digest
+                else []
+            )
         )
         .select(cs.ends_with("_a"))
         .rename(lambda column_name: column_name.removesuffix("_a"))
-        .unnest(["roc_auc_tdigest"] if use_t_digest else [])
+        .unnest(["roc_auc_tdigest", 'calibration_tdigest'] if use_t_digest else [])
         .sort(grp_by, descending=False)
     )
     if drop_fpr_tpr:
         auc_data = (
             auc_data
             .drop(
-                cs.by_name(['fpr', 'tpr', 'precision', 'recall'], require_all=False),
+                cs.by_name(
+                    ['fpr', 'tpr', 'precision', 'recall', 'calibration_bin', 'calibration_proba', 'calibration_rate'],
+                    require_all=False),
                 strict=False
             )
         )
@@ -276,6 +320,9 @@ def calculate_engagement_scores(
     if isinstance(ih_analysis, pd.DataFrame):
         ih_analysis = pl.from_pandas(ih_analysis)
 
+    if isinstance(ih_analysis, pl.DataFrame):
+        ih_analysis = ih_analysis.clone()
+
     sort_list = []
     if "x" in config.keys():
         sort_list.append(config["x"])
@@ -289,7 +336,13 @@ def calculate_engagement_scores(
     data_copy = ih_analysis.filter(pl.col("Negatives") > 0)
     column_name_map = {"z_score": "Lift_Z_Score", "z_p_val": "Lift_P_Val"}
 
-    grp_by = config["group_by"]
+    grp_by = config.get("group_by", None)
+    if not grp_by:
+        grp_by = ['ALL']
+        data_copy = data_copy.with_columns(
+            pl.lit('ALL').alias('ALL')
+        )
+    ret_dtype = pl.Struct({"z_score": pl.Float64, "z_p_val": pl.Float64})
     if grp_by:
         data_copy = (
             data_copy.with_columns(
@@ -361,7 +414,8 @@ def calculate_engagement_scores(
                         x["Negatives_Test"],
                         x["Negatives_Control"],
                     ),
-                    return_dtype=pl.Struct,
+                    return_dtype=ret_dtype,
+                    returns_scalar=True
                 )
                 .alias("Z_Test")
             )
@@ -377,7 +431,7 @@ def calculate_engagement_scores(
                 ]
             )
             .rename(lambda column_name: column_name_map.get(column_name, column_name))
-            .sort(config["group_by"], descending=False)
+            .sort(grp_by, descending=False)
         )
 
     return data_copy
@@ -388,19 +442,25 @@ def calculate_conversion_scores(
 ) -> pl.DataFrame:
     if isinstance(ih_analysis, pd.DataFrame):
         ih_analysis = pl.from_pandas(ih_analysis)
+    if isinstance(ih_analysis, pl.DataFrame):
+        ih_analysis = ih_analysis.clone()
 
     grp_by = config["group_by"]
     metric = config["metric"]
     copy_data = (
         ih_analysis.group_by(grp_by)
         .agg(
-            pl.sum("Negatives").alias("Negatives"),
-            pl.sum("Positives").alias("Positives"),
-            pl.sum("Count").alias("Count"),
-            pl.sum("Revenue").alias("Revenue"),
+            [
+                pl.col("Negatives").sum(),
+                pl.col("Positives").sum(),
+                pl.col("Count").sum(),
+                pl.col("Revenue").sum(),
+                pl.col('Touchpoints').sum()
+            ]
         )
         .with_columns(
-            [(pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives"))).alias("ConversionRate")]
+            [(pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives"))).alias("ConversionRate"),
+             (pl.col("Touchpoints") / pl.col("Positives")).alias("AvgTouchpoints")]
         )
         .with_columns(
             [
@@ -427,6 +487,8 @@ def group_engagement_data(
 ) -> pl.DataFrame:
     if isinstance(eng_data, pd.DataFrame):
         eng_data = pl.from_pandas(eng_data)
+    if isinstance(eng_data, pl.DataFrame):
+        eng_data = eng_data.clone()
 
     grp_by = config["group_by"] + get_config()["metrics"]["global_filters"]
     grp_by = list(set(grp_by))
@@ -448,6 +510,8 @@ def group_conversion_data(
 ) -> pl.DataFrame:
     if isinstance(conv_data, pd.DataFrame):
         conv_data = pl.from_pandas(conv_data)
+    if isinstance(conv_data, pl.DataFrame):
+        conv_data = conv_data.clone()
 
     data_copy = conv_data.filter(pl.col("Negatives") > 0)
 
@@ -470,6 +534,8 @@ def group_descriptive_data(
 ) -> pl.DataFrame:
     if isinstance(data, pd.DataFrame):
         data = pl.from_pandas(data)
+    if isinstance(data, pl.DataFrame):
+        data = data.clone()
     return data
 
 
@@ -478,12 +544,14 @@ def calculate_descriptive_scores(
 ) -> pl.DataFrame:
     if isinstance(data, pd.DataFrame):
         data = pl.from_pandas(data)
+    if isinstance(data, pl.DataFrame):
+        data = data.clone()
 
     m_config = get_config()["metrics"][config["metric"]]
     use_t_digest = (
         strtobool(m_config["use_t_digest"])
         if "use_t_digest" in m_config.keys()
-        else False
+        else True
     )
     logger.debug("Use t-digest for scores: " + str(use_t_digest))
     columns_conf = m_config["columns"]
@@ -704,12 +772,52 @@ def calculate_descriptive_scores(
 
 
 def calculate_clv_scores(
-        exp_data: Union[pl.DataFrame, pd.DataFrame], config: dict
+        data: Union[pl.DataFrame, pd.DataFrame], config: dict
 ) -> pl.DataFrame:
-    if isinstance(exp_data, pd.DataFrame):
-        exp_data = pl.from_pandas(exp_data)
+    if isinstance(data, pd.DataFrame):
+        data = pl.from_pandas(data)
+    if isinstance(data, pl.DataFrame):
+        data = data.clone()
     m_config = get_config()["metrics"][config["metric"]]
-    totals_frame = rfm_summary(exp_data, m_config)
-    if "x" in config.keys():
-        totals_frame = totals_frame.sort(config["x"], descending=True)
+    totals_frame = rfm_summary(data, m_config)
     return totals_frame
+
+
+def merge_descriptive_digests(
+        data: Union[pl.DataFrame, pd.DataFrame], config: dict) -> pl.DataFrame:
+    if isinstance(data, pd.DataFrame):
+        data = pl.from_pandas(data)
+    if isinstance(data, pl.DataFrame):
+        data = data.clone()
+
+    m_config = get_config()["metrics"][config["metric"]]
+    use_t_digest = (
+        strtobool(m_config["use_t_digest"])
+        if "use_t_digest" in m_config.keys()
+        else True
+    )
+    logger.debug("Use t-digest for scores: " + str(use_t_digest))
+    columns_conf = m_config["columns"]
+    num_columns = [col for col in columns_conf if (col + "_Mean") in data.columns]
+    grp_by = config["group_by"]
+
+    cdata = data.clone()
+    if not use_t_digest:
+        return pl.DataFrame()
+    else:
+        tdigest_aggs = [
+            pl.map_groups(
+                exprs=[pl.col(f'{c}_tdigest')],
+                function=lambda s: merge_digests(s),
+                return_dtype=pl.Binary,
+                returns_scalar=True
+            ).alias(f'{c}_tdigest_a') for c in num_columns
+        ]
+
+        cdata = cdata.group_by(grp_by).agg([pl.col(grp_by).first().name.suffix("_a")] + tdigest_aggs)
+        cdata = (
+            cdata.select(cs.ends_with("_a"))
+            .rename(lambda column_name: column_name.removesuffix("_a"))
+            .sort(grp_by, descending=False)
+        )
+    return cdata
